from pathlib import Path
from subprocess import run


def test_bakta_mock_skipped_features(tmpdir):
    # fast test skipping all feature detections
    proc = run(["bin/bakta", '--db', 'test/db', '--output', tmpdir, '--prefix', 'test', '--skip-tmrna', '--skip-trna', '--skip-rrna', '--skip-ncrna', '--skip-ncrna-region', '--skip-crispr', '--skip-cds', '--skip-sorf', '--skip-ori', '--skip-gap', 'test/data/NC_002127.1.fna'])
    assert proc.returncode == 0
    assert Path.exists(Path(tmpdir).joinpath('test.log'))
    assert Path.exists(Path(tmpdir).joinpath('test.json'))
    assert Path.exists(Path(tmpdir).joinpath('test.tsv'))
    assert Path.exists(Path(tmpdir).joinpath('test.gff3'))
    assert Path.exists(Path(tmpdir).joinpath('test.gbff'))
    assert Path.exists(Path(tmpdir).joinpath('test.fna'))
    assert Path.exists(Path(tmpdir).joinpath('test.faa'))


def test_bakta_plasmid(tmpdir):
    # full test on plasmid
    proc = run(["bin/bakta", '--db', 'test/db', '--output', tmpdir, '--prefix', 'test', 'test/data/NC_002127.1.fna'])
    assert proc.returncode == 0
    assert Path.exists(Path(tmpdir).joinpath('test.log'))
    assert Path.exists(Path(tmpdir).joinpath('test.json'))
    assert Path.exists(Path(tmpdir).joinpath('test.tsv'))
    assert Path.exists(Path(tmpdir).joinpath('test.gff3'))
    assert Path.exists(Path(tmpdir).joinpath('test.gbff'))
    assert Path.exists(Path(tmpdir).joinpath('test.fna'))
    assert Path.exists(Path(tmpdir).joinpath('test.faa'))


def test_bakta_genome(tmpdir):
    # full test on plasmid
    proc = run(["bin/bakta", '--db', 'test/db', '--output', tmpdir, '--prefix', 'test', 'test/data/GCF_000008865.2.fna.gz'])
    assert proc.returncode == 0
    assert Path.exists(Path(tmpdir).joinpath('test.log'))
    assert Path.exists(Path(tmpdir).joinpath('test.json'))
    assert Path.exists(Path(tmpdir).joinpath('test.tsv'))
    assert Path.exists(Path(tmpdir).joinpath('test.gff3'))
    assert Path.exists(Path(tmpdir).joinpath('test.gbff'))
    assert Path.exists(Path(tmpdir).joinpath('test.fna'))
    assert Path.exists(Path(tmpdir).joinpath('test.faa'))



import logging
import subprocess as sp
from collections import OrderedDict

from Bio.Seq import Seq

import bakta.config as cfg
import bakta.constants as bc
import bakta.utils as bu
import bakta.psc as psc

log = logging.getLogger('S_ORF')


def extract(genome):
    """Predict open reading frames in mem via BioPython."""

    orfs = []
    for contig in genome['contigs']:
        dna_seq = Seq(contig['sequence'])
        for strand, seq in [(bc.STRAND_FORWARD, dna_seq), (bc.STRAND_REVERSE, dna_seq.reverse_complement())]:  # strands +/-
            for frame in range(3):  # frames 1/2/3 -> 0, 1, 2
                seq_frame = seq[frame:]

                # remove non-triplet tail nucleotides
                residue = len(seq_frame) % 3
                if(residue != 0):
                    seq_frame = seq_frame[:-residue]

                aa_seq = str(seq_frame.translate(table=cfg.translation_table, stop_symbol='*', to_stop=False, cds=False))
                aa_start = aa_seq.find('M')
                aa_end = aa_seq.find('*', aa_start)
                while aa_start > -1 and aa_end > -1:
                    orf_length = aa_end - aa_start
                    if(orf_length >= bc.MIN_SORF_LENGTH and orf_length <= bc.MAX_SORF_LENGTH):  # get all CDS starts (M)
                        if(strand == bc.STRAND_FORWARD):
                            dna_start = aa_start * 3 + frame + 1  # +1: 0 based idx to 1 based
                            dna_stop = aa_end * 3 + 2 + frame + 1
                        else:
                            dna_start = len(seq) - frame - (aa_end + 1) * 3 + 1
                            dna_stop = len(seq) - frame - aa_start * 3
                        sequence = aa_seq[aa_start:aa_end]
                        (aa_digest, aa_hexdigest) = bu.calc_aa_hash(sequence)

                        sorf = OrderedDict()
                        sorf['type'] = bc.FEATURE_SORF
                        sorf['contig'] = contig['id']
                        sorf['start'] = dna_start
                        sorf['stop'] = dna_stop
                        sorf['strand'] = strand
                        sorf['product'] = None
                        sorf['gene'] = None
                        sorf['frame'] = frame + 1
                        sorf['db_xrefs'] = []
                        sorf['sequence'] = sequence
                        sorf['aa_digest'] = aa_digest
                        sorf['aa_hexdigest'] = aa_hexdigest
                        
                        orfs.append(sorf)
                        log.debug(
                            'contig=%s, start=%i, stop=%i, strand=%s, frame=%i, length=%i, seq=%s',
                            contig['id'], sorf['start'], sorf['stop'], strand, frame, len(sequence), sequence
                        )
                    aa_start = aa_seq.find('M', aa_start + 1)
                    if(aa_start > aa_end):
                        aa_end = aa_seq.find('*', aa_start)

    log.info('predicted=%i', len(orfs))
    return orfs


def get_feature_start(feature):
    return feature['start'] if feature['strand'] == '+' else feature['stop']


def get_feature_stop(feature):
    return feature['stop'] if feature['strand'] == '+' else feature['start']


def overlap_filter(genome, orfs_raw):
    """Filter in-mem ORFs by overlapping CDSs."""

    contig_t_rnas = {k['id']: [] for k in genome['contigs']}
    for t_rna in genome['features'].get(bc.FEATURE_T_RNA, []):
        t_rnas = contig_t_rnas[t_rna['contig']]
        t_rnas.append(t_rna)
    for tm_rna in genome['features'].get(bc.FEATURE_TM_RNA, []):
        t_rnas = contig_t_rnas[tm_rna['contig']]
        t_rnas.append(tm_rna)

    contig_r_rnas = {k['id']: [] for k in genome['contigs']}
    for r_rna in genome['features'].get(bc.FEATURE_R_RNA, []):
        r_rnas = contig_r_rnas[r_rna['contig']]
        r_rnas.append(r_rna)

    # contig_nc_rnas = {k['id']: [] for k in genome['contigs']}
    # for nc_rna in genome['features'].get(bc.FEATURE_NC_RNA, []):
    #     nc_rnas = contig_nc_rnas[nc_rna['contig']]
    #     nc_rnas.append(nc_rna)
    # for nc_rna in genome['features'].get(bc.FEATURE_NC_RNA_REGION, []):
    #     nc_rnas = contig_nc_rnas[nc_rna['contig']]
    #     nc_rnas.append(nc_rna)

    contig_crispr_arrays = {k['id']: [] for k in genome['contigs']}
    for crispr_array in genome['features'].get(bc.FEATURE_CRISPR, []):
        crispr_arrays = contig_crispr_arrays[crispr_array['contig']]
        crispr_arrays.append(crispr_array)

    contig_cdss = {k['id']: [] for k in genome['contigs']}
    for cds in genome['features'].get(bc.FEATURE_CDS, []):
        cdss = contig_cdss[cds['contig']]
        cdss.append(cds)

    contig_sorfs = {k['id']: [] for k in genome['contigs']}
    for sorf in orfs_raw:
        orfs = contig_sorfs[sorf['contig']]
        orfs.append(sorf)

    discarded_sorfs = {}
    for contig in genome['contigs']:
        log.debug('filter short ORFs on contig: %s', contig['id'])
        sorfs = contig_sorfs[contig['id']]

        # filter CDS overlapping ORFs
        for sorf in sorfs:
            for cds in contig_cdss[contig['id']]:
                # log.debug('filter short ORFs by CDS: %s%i[%i->%i]', cds['strand'], cds['frame'], cds['start'], cds['stop'])
                if(sorf['strand'] == cds['strand']):
                    if(sorf['frame'] == cds['frame']):
                        # fast/simple overlap detection for in frame orientation
                        if(cds['stop'] < sorf['start'] or cds['start'] > sorf['stop']):
                            continue
                        else:
                            key = f"{sorf['contig']}-{sorf['start']}-{sorf['stop']}-{sorf['strand']}-{sorf['aa_hexdigest']}"
                            discarded_sorfs[key] = sorf
                    else:
                        if(sorf['start'] < cds['start'] and sorf['stop'] > cds['start']):
                            # out-of-frame sorf partially overlapping CDS upstream
                            # ToDo: add max overlap threshold
                            continue
                        elif(sorf['start'] >= cds['start'] and sorf['stop'] <= cds['stop']):
                            # out-of-frame sorf completely overlapped by CDS
                            key = f"{sorf['contig']}-{sorf['start']}-{sorf['stop']}-{sorf['strand']}-{sorf['aa_hexdigest']}"
                            discarded_sorfs[key] = sorf
                        elif(sorf['start'] < cds['stop'] and sorf['stop'] > cds['stop']):
                            # out-of-frame sorf partially overlapping CDS downstream
                            # ToDo: add max overlap threshold
                            continue
                else:
                    if(sorf['frame'] == cds['frame']):
                        # fast/simple overlap detection for in frame orientation
                        if(cds['stop'] < sorf['start'] or cds['start'] > sorf['stop']):
                            continue
                        else:
                            key = f"{sorf['contig']}-{sorf['start']}-{sorf['stop']}-{sorf['strand']}-{sorf['aa_hexdigest']}"
                            discarded_sorfs[key] = sorf
                    else:
                        # ToDo: add out-of-frame filters
                        if(sorf['start'] < cds['start'] and sorf['stop'] > cds['start']):
                            # out-frame sorf partially overlapping CDS upstream
                            # ToDo: add max overlap threshold
                            continue
                        elif(sorf['start'] >= cds['start'] and sorf['stop'] <= cds['stop']):
                            # out-frame sorf completely overlapped by CDS
                            key = f"{sorf['contig']}-{sorf['start']}-{sorf['stop']}-{sorf['strand']}-{sorf['aa_hexdigest']}"
                            discarded_sorfs[key] = sorf
                        elif(sorf['start'] < cds['stop'] and sorf['stop'] > cds['stop']):
                            # out-frame sorf partially overlapping CDS downstream
                            # ToDo: add max overlap threshold
                            continue

            # filter rRNA overlapping ORFs
            for r_rna in contig_r_rnas[contig['id']]:
                # log.debug('filter short ORFs by rRNA: %s[%i->%i]', r_rna['strand'], r_rna['start'], r_rna['stop'])
                # fast/simple overlap detection for rRNAs
                if(sorf['stop'] < r_rna['start'] or sorf['start'] > r_rna['stop']):
                    continue
                else:
                    key = f"{sorf['contig']}-{sorf['start']}-{sorf['stop']}-{sorf['strand']}-{sorf['aa_hexdigest']}"
                    discarded_sorfs[key] = sorf

            # filter tRNA overlapping ORFs
            # log.debug('filter short ORFs by tRNA: %s[%i->%i]', t_rna['strand'], t_rna['start'], t_rna['stop'])
            for t_rna in contig_t_rnas[contig['id']]:
                # fast/simple overlap detection for tRNAs
                if(sorf['stop'] < t_rna['start'] or sorf['start'] > t_rna['stop']):
                    continue
                else:
                    key = f"{sorf['contig']}-{sorf['start']}-{sorf['stop']}-{sorf['strand']}-{sorf['aa_hexdigest']}"
                    discarded_sorfs[key] = sorf
        
            # filter CRISPR array overlapping ORFs
            # log.debug('filter short ORFs by CRISPR: [%i->%i]', crispr['start'], crispr['stop'])
            for crispr in contig_crispr_arrays[contig['id']]:
                # fast/simple overlap detection for CRISPR
                if(sorf['stop'] < crispr['start'] or sorf['start'] > crispr['stop']):
                    continue
                else:
                    key = f"{sorf['contig']}-{sorf['start']}-{sorf['stop']}-{sorf['strand']}-{sorf['aa_hexdigest']}"
                    discarded_sorfs[key] = sorf
    
    valid_sorfs = []
    for sorfs in contig_sorfs.values():
        for sorf in sorfs:
            key = f"{sorf['contig']}-{sorf['start']}-{sorf['stop']}-{sorf['strand']}-{sorf['aa_hexdigest']}"
            if(key not in discarded_sorfs):
                valid_sorfs.append(sorf)

    log.info('sORF filter: # valid=%i, # discarded=%i', len(valid_sorfs), len(discarded_sorfs.values()))
    return valid_sorfs, discarded_sorfs


def annotation_filter(sorfs):
    """Filter sORFs according to available annotations."""
    valid_sorfs = []
    for sorf in sorfs:
        gene = None
        product = None
        
        ips = sorf.get('ips', None)
        if(ips is not None):
            tmp = ips.get('gene', '')
            if(tmp != ''):
                gene = tmp
            tmp = ips.get('product', '')
            if(tmp != ''):
                product = tmp
        
        psc = sorf.get('psc', None)
        if(psc is not None):
            tmp = psc.get('gene', '')
            if(tmp != ''):
                gene = tmp
            tmp = psc.get('product', '')
            if(tmp != ''):
                product = tmp

        if(gene is None and product is None):
            sorf['hypothetical'] = True
        else:
            valid_sorfs.append(sorf)
    
    return valid_sorfs
    

def search_pscs(sorfs):
    """Conduct homology search of sORFs against sORF db."""
    sorf_fasta_path = cfg.tmp_path.joinpath('sorf.faa')
    with sorf_fasta_path.open(mode='w') as fh:
        for sorf in sorfs:
            fh.write(f">{sorf['aa_hexdigest']}\n{sorf['sequence']}\n")
    diamond_output_path = cfg.tmp_path.joinpath('diamond.sorf.tsv')
    diamond_db_path = cfg.db_path.joinpath('sorf.dmnd')
    cmd = [
        'diamond',
        'blastp',
        '--db', str(diamond_db_path),
        '--query', str(sorf_fasta_path),
        '--out', str(diamond_output_path),
        '--id', str(int(bc.MIN_PROTEIN_IDENTITY * 100)),  # '90',
        '--query-cover', str(int(bc.MIN_SORF_COVERAGE * 100)),  # '90'
        '--subject-cover', str(int(bc.MIN_SORF_COVERAGE * 100)),  # '90'
        '--max-target-seqs', '1',  # single best output
        '--outfmt', '6',
        '--threads', str(cfg.threads),
        '--tmpdir', str(cfg.tmp_path),  # use tmp folder
        '--block-size', '3'  # slightly increase block size for faster executions
    ]
    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if(proc.returncode != 0):
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('sORF failed! diamond-error-code=%d', proc.returncode)
        raise Exception(f'diamond error! error code: {proc.returncode}')

    sorf_by_aa_digest = {sorf['aa_hexdigest']: sorf for sorf in sorfs}
    with diamond_output_path.open() as fh:
        for line in fh:
            (sorf_hash, cluster_id, identity, alignment_length, align_mismatches,
                align_gaps, query_start, query_end, subject_start, subject_end,
                evalue, bitscore) = line.split('\t')
            sorf = sorf_by_aa_digest[sorf_hash]
            query_cov = int(alignment_length) / len(sorf['sequence'])
            identity = float(identity) / 100
            if(query_cov >= bc.MIN_PROTEIN_COVERAGE and identity >= bc.MIN_PROTEIN_IDENTITY):
                sorf['psc'] = {
                    psc.DB_PSC_COL_UNIREF90: cluster_id,
                    'query-cov': query_cov,
                    'identity': identity
                }
                log.info(
                    'homology: contig=%s, start=%i, stop=%i, strand=%s, aa-length=%i, query-cov=%0.3f, identity=%0.3f, UniRef90=%s',
                    sorf['contig'], sorf['start'], sorf['stop'], sorf['strand'], len(sorf['sequence']), query_cov, identity, cluster_id
                )

    pscs_found = []
    pscs_not_found = []
    for sorf in sorfs:
        if('psc' in sorf):
            pscs_found.append(sorf)
        else:
            pscs_not_found.append(sorf)
    log.info('found=%i', len(pscs_found))
    return pscs_found, pscs_not_found




import logging

import bakta
import bakta.config as cfg
import bakta.constants as bc

log = logging.getLogger('TSV')

############################################################################
# Inference terms
# 
# tRNA: 'tRNAscan'
# tmRNA: 'aragorn'
# rRNA: 'Rfam:%s' % subject_id
# ncRNA genes: 'Rfam:%s' % subject_id
# ncRNA regions: 'Rfam:%s' % subject_id
# CDS hyp: 'Prodigal'
# CDS PSC: 'UniProtKB:%s' % psc[DB_PSC_COL_UNIREF90]
# CDS IPS: 'UniProtKB:%s' % psc[DB_PSC_COL_UNIREF100]
# CDS sORF: 'similar to AA sequence:UniProtKB:%s' % psc[DB_PSC_COL_UNIREF100]
############################################################################


def write_tsv(contigs, features_by_contig, tsv_path):
    """Export features in TSV format."""
    log.info('write tsv output: path=%s', tsv_path)
    
    with tsv_path.open('w') as fh:
        fh.write(f'#Annotated with Bakta v{bakta.__version__}, https://github.com/oschwengers/bakta\n')
        fh.write(f"#Database v{cfg.db_info['major']}.{cfg.db_info['minor']}, https://doi.org/10.5281/zenodo.4247252\n")
        fh.write('#Sequence Id\tType\tStart\tStop\tStrand\tGene\tProduct\tDbXrefs\n')
        for contig in contigs:
            for feat in features_by_contig[contig['id']]:
                feat_type = feat['type']
                if(feat['type'] == bc.FEATURE_GAP):
                    feat_type = bc.INSDC_FEATURE_ASSEMBLY_GAP if feat['length'] >= 100 else bc.INSDC_FEATURE_GAP
                
                gene = feat['gene'] if feat.get('gene', None) else ''
                fh.write('\t'.join([feat['contig'], feat_type, str(feat['start']), str(feat['stop']), feat['strand'], gene, feat.get('product', ''), ', '.join(sorted(feat.get('db_xrefs', [])))]))
                fh.write('\n')
    return



import logging
from datetime import date

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation, CompoundLocation, AfterPosition, BeforePosition

import bakta
import bakta.config as cfg
import bakta.constants as bc

log = logging.getLogger('GENBANK')


def write_genbank(genome, features, genbank_path):
    log.info('write GenBank output: path=%s', genbank_path)

    contig_list = []
    for contig in genome['contigs']:
        contig_annotations = {
            'molecule_type': 'DNA',
            'source': genome['taxon'],
            'date': date.today().strftime('%d-%b-%Y').upper(),
            'topology': contig['topology'],
            'data_file_division': 'HGT' if contig['type'] == bc.REPLICON_CONTIG else 'BCT',
            'comment': f"Annotated with Bakta v{bakta.__version__}, https://github.com/oschwengers/bakta\nDatabase v{cfg.db_info['major']}.{cfg.db_info['minor']}, https://doi.org/10.5281/zenodo.4247252"
            # TODO: taxonomy
        }
        source_qualifiers = {
            'mol_type': 'genomic DNA'
        }

        description = ''
        if(genome['taxon']):
            contig_annotations['organism'] = genome['taxon']
            source_qualifiers['organism'] = genome['taxon']
            description = genome['taxon']
        if(genome['strain']):
            source_qualifiers['strain'] = genome['strain']
        
        if(contig['type'] == bc.REPLICON_PLASMID):
            source_qualifiers['plasmid'] = contig['name'] if contig.get('name', None) else 'unnamed'
            description = f"{description} plasmid {contig.get('name', 'unnamed')}"
            description += ', complete sequence' if contig['complete'] else ', whole genome shotgun sequence'
        elif(contig['type'] == bc.REPLICON_CHROMOSOME):
            source_qualifiers['chromosome'] = contig['name'] if contig.get('name', None) else contig['id']
            description = f'{description} chromosome, complete genome' if contig['complete'] else f"{description} chromosome {contig['id']}, whole genome shotgun sequence"
        else:
            description += f" {contig['id']}, whole genome shotgun sequence"
        
        if(len(description) > 0 and description[0] == ' '):  # discard potential leading whitespace
            description = description[1:]

        contig_rec = SeqIO.SeqRecord(id='.', name=contig['id'], description=description, annotations=contig_annotations, seq=Seq(contig['sequence']))
        
        source = SeqFeature(FeatureLocation(0, contig['length'], strand=+1), type='source', qualifiers=source_qualifiers)
        seq_feature_list = [source]

        for feature in features:
            if(feature['contig'] == contig['id']):
                insdc_feature_type = None
                qualifiers = {}
                if('db_xrefs' in feature):
                    qualifiers['db_xref'] = feature['db_xrefs']
                if('product' in feature):
                    qualifiers['product'] = feature['product']
                if('locus' in feature):
                    qualifiers['locus_tag'] = feature['locus']
                
                if(feature['type'] == bc.FEATURE_GAP):
                    insdc_feature_type = bc.INSDC_FEATURE_GAP
                    qualifiers['estimated_length'] = feature['length']
                elif(feature['type'] == bc.FEATURE_ORIC or feature['type'] == bc.FEATURE_ORIV):
                    # TODO: Add fuzzy positions for oriC/oriV
                    insdc_feature_type = bc.INSDC_FEATURE_ORIGIN_REPLICTION
                    qualifiers['inference'] = 'similar to DNA sequence'
                elif(feature['type'] == feature['type'] == bc.FEATURE_ORIT):
                    # TODO: Add fuzzy positions for oriT
                    insdc_feature_type = bc.INSDC_FEATURE_ORIGIN_TRANSFER
                    qualifiers['inference'] = 'similar to DNA sequence'
                elif(feature['type'] == bc.FEATURE_CDS) or (feature['type'] == bc.FEATURE_SORF):
                    qualifiers['translation'] = feature['sequence']
                    qualifiers['codon_start'] = 1
                    qualifiers['transl_table'] = cfg.translation_table
                    insdc_feature_type = bc.INSDC_FEATURE_CDS
                    inference = []
                    inference.append('ab initio prediction:Prodigal:2.6' if feature['type'] == bc.FEATURE_CDS else 'ab initio prediction:Bakta')
                    if('hypothetical' not in feature):
                        if('ups' in feature):
                            if('ncbi_nrp_id' in feature['ups']):
                                qualifiers['protein_id'] = feature['ups']['ncbi_nrp_id']
                        if('ips' in feature):
                            if('uniref100_id' in feature['ips']):
                                ips_subject_id = feature['ips']['uniref100_id']
                                inference.append(f'similar to AA sequence:UniProtKB:{ips_subject_id}')
                        if('psc' in feature):
                            if('uniref90_id' in feature['psc']):
                                psc_subject_id = feature['psc']['uniref90_id']
                                inference.append(f'similar to AA sequence:UniProtKB:{psc_subject_id}')
                    qualifiers['inference'] = inference
                elif(feature['type'] == bc.FEATURE_T_RNA):
                    # TODO: Position anticodon
                    if('notes' in feature):
                        if('anti_codon' in feature):
                            qualifiers['note'] = feature['notes']
                            i = feature['product'].find('-')
                            t_rna_type = feature['product'][i+1:].lower()
                            qualifiers['anticodon'] = f"(aa:{t_rna_type},seq:{feature['anti_codon'].lower()})"
                    qualifiers['inference'] = 'profile:tRNAscan:2.0'
                    insdc_feature_type = bc.INSDC_FEATURE_T_RNA
                elif(feature['type'] == bc.FEATURE_TM_RNA):
                    qualifiers['inference'] = 'profile:aragorn:1.2'
                    insdc_feature_type = bc.INSDC_FEATURE_TM_RNA
                elif(feature['type'] == bc.FEATURE_R_RNA):
                    for dbxref in feature['db_xrefs']:
                        if(dbxref.split(':')[0] == 'RFAM'):
                            rfam_id = dbxref.split(':')[1]
                            qualifiers['inference'] = f'profile:Rfam:{rfam_id}'
                    insdc_feature_type = bc.INSDC_FEATURE_R_RNA
                elif(feature['type'] == bc.FEATURE_NC_RNA):
                    # TODO: ncRNA_class
                    for dbxref in feature['db_xrefs']:
                        if(dbxref.split(':')[0] == 'RFAM'):
                            rfam_id = dbxref.split(':')[1]
                            qualifiers['inference'] = f'profile:Rfam:{rfam_id}'
                    qualifiers['ncRNA_class'] = 'other'
                    insdc_feature_type = bc.INSDC_FEATURE_NC_RNA
                elif(feature['type'] == bc.FEATURE_NC_RNA_REGION):
                    for dbxref in feature['db_xrefs']:
                        if(dbxref.split(':')[0] == 'RFAM'):
                            rfam_id = dbxref.split(':')[1]
                            qualifiers['inference'] = f'profile:Rfam:{rfam_id}'
                    qualifiers['ncRNA_class'] = 'other'
                    insdc_feature_type = bc.INSDC_FEATURE_REGULATORY
                elif(feature['type'] == bc.FEATURE_CRISPR):
                    qualifiers['repeats'] = feature['repeats']
                    qualifiers['repeat_consensus'] = feature['repeat_consensus']
                    qualifiers['repeat_length'] = feature['repeat_length']
                    qualifiers['spacer_length'] = feature['spacer_length']
                    insdc_feature_type = bc.INSDC_FEATURE_MISC_FEATURE
                
                strand = None
                if(feature['strand'] == bc.STRAND_FORWARD):
                    strand = 1
                elif(feature['strand'] == bc.STRAND_REVERSE):
                    strand = -1
                elif(feature['strand'] == bc.STRAND_UNKNOWN):
                    strand = 0
                
                start = feature['start'] - 1
                stop = feature['stop']
                if('edge' in feature):
                    fl_1 = FeatureLocation(start, contig['length'], strand=strand)
                    fl_2 = FeatureLocation(0, stop, strand=strand)
                    feature_location = CompoundLocation([fl_1, fl_2])
                else:
                    if('truncated' in feature):
                        if(feature['truncated'] == bc.FEATURE_END_5_PRIME):
                            if(feature['strand'] == bc.STRAND_FORWARD):
                                start = BeforePosition(start)
                            else:
                                stop = AfterPosition(stop)
                        elif(feature['truncated'] == bc.FEATURE_END_3_PRIME):
                            if(feature['strand'] == bc.STRAND_FORWARD):
                                stop = AfterPosition(stop)
                            else:
                                start = BeforePosition(start)
                        else:
                            start = BeforePosition(start)
                            stop = AfterPosition(stop)
                    feature_location = FeatureLocation(start, stop, strand=strand)

                if(feature.get('locus', None)):
                    gene_qualifier = {
                        'locus_tag': feature['locus']
                    }
                    if(feature.get('gene', None)):
                        qualifiers['gene'] = feature['gene']
                        gene_qualifier['gene'] = feature['gene']
                    gen_seqfeat = SeqFeature(feature_location, type='gene', qualifiers=gene_qualifier)
                    seq_feature_list.append(gen_seqfeat)
                feat_seqfeat = SeqFeature(feature_location, type=insdc_feature_type, qualifiers=qualifiers)
                seq_feature_list.append(feat_seqfeat)
        contig_rec.features = seq_feature_list
        contig_list.append(contig_rec)

    with genbank_path.open('w', encoding='utf-8') as fh:
        SeqIO.write(contig_list, fh, format='genbank')

    return




import logging

import bakta.constants as bc

log = logging.getLogger('ANNOTATION')


def combine_ips_psc_annotation(feature):
    ups = feature.get('ups', None)
    ips = feature.get('ips', None)
    psc = feature.get('psc', None)

    gene = None
    product = None
    db_xrefs = set()
    if(psc):
        psc_gene = psc.get('gene', None)
        if(psc_gene):
            gene = psc_gene
        psc_product = psc.get('product', None)
        if(psc_product):
            product = psc_product
        for db_xref in psc['db_xrefs']:
            db_xrefs.add(db_xref)
    if(ups):
        for db_xref in ups['db_xrefs']:
            db_xrefs.add(db_xref)
    if(ips):
        ips_gene = ips.get('gene', None)
        if(ips_gene):
            gene = ips_gene
        ips_product = ips.get('product', None)
        if(ips_product):
            product = ips_product
        for db_xref in ips['db_xrefs']:
            db_xrefs.add(db_xref)
    
    if(gene):
        feature['gene'] = gene
    if(product):
        feature['product'] = product
    else:
        feature['product'] = bc.HYPOTHETICAL_PROTEIN
        feature['hypothetical'] = True
    feature['db_xrefs'] = sorted(list(db_xrefs))


def detect_feature_overlaps(genome):
    """Apply feature type specific hierarchical feature overlap filters.
    tRNA < tmRNA
    CDS < tmRNA, tRNA, rRNA, CRISPR
    sORF < mRNA, tRNA, rRNA, CRISPR, CDS (in-frame & entirely overlapping), sORF (shorter, weaker annotations)
    """
    contig_t_rnas = {k['id']: [] for k in genome['contigs']}
    for t_rna in genome['features'].get(bc.FEATURE_T_RNA, []):
        t_rnas = contig_t_rnas[t_rna['contig']]
        t_rnas.append(t_rna)
    contig_tm_rnas = {k['id']: [] for k in genome['contigs']}
    for tm_rna in genome['features'].get(bc.FEATURE_TM_RNA, []):
        tm_rnas = contig_tm_rnas[tm_rna['contig']]
        tm_rnas.append(tm_rna)
    contig_r_rnas = {k['id']: [] for k in genome['contigs']}
    for r_rna in genome['features'].get(bc.FEATURE_R_RNA, []):
        r_rnas = contig_r_rnas[r_rna['contig']]
        r_rnas.append(r_rna)
    contig_crispr_arrays = {k['id']: [] for k in genome['contigs']}
    for crispr_array in genome['features'].get(bc.FEATURE_CRISPR, []):
        crispr_arrays = contig_crispr_arrays[crispr_array['contig']]
        crispr_arrays.append(crispr_array)
    contig_cdss = {k['id']: [] for k in genome['contigs']}
    for cds in genome['features'].get(bc.FEATURE_CDS, []):
        cdss = contig_cdss[cds['contig']]
        cdss.append(cds)
    contig_sorfs = {k['id']: [] for k in genome['contigs']}
    for sorf in genome['features'].get(bc.FEATURE_SORF, []):
        sorfs = contig_sorfs[sorf['contig']]
        sorfs.append(sorf)
    
    for contig in genome['contigs']:  # find feature overlaps contig-wise to increase the performance
        log.debug('filter features on contig: %s', contig['id'])

        # mark tRNAs overlapping with tmRNAs
        for tRNA in contig_t_rnas[contig['id']]:
            for tmRNA in contig_tm_rnas[contig['id']]:
                if(tRNA['stop'] < tmRNA['start'] or tRNA['start'] > tmRNA['stop']):
                    continue
                else:  # overlap -> remove tRNA
                    overlap = f"[{max(tRNA['start'], tmRNA['start'])},{min(tRNA['stop'], tmRNA['stop'])}]"
                    tRNA['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_TM_RNA,
                        'description': f"{bc.FEATURE_TM_RNA} overlap with ({tmRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: tRNA (%s) [%i, %i] overlapping with tmRNA (%s) [%i, %i] at %s on contig=%s",
                        tRNA['product'], tRNA['start'], tRNA['stop'], tmRNA['product'], tmRNA['start'], tmRNA['stop'], overlap, tRNA['contig']
                    )
        
        # mark CDS overlapping with tRNAs, tmRNAs, rRNAs, CRISPRs
        for cds in contig_cdss[contig['id']]:
            # tmRNA overlaps
            for tmRNA in contig_tm_rnas[contig['id']]:
                if(cds['stop'] < tmRNA['start'] or cds['start'] > tmRNA['stop']):
                    continue
                else:  # overlap -> remove cds
                    overlap = f"[{max(cds['start'], tmRNA['start'])},{min(cds['stop'], tmRNA['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_TM_RNA,
                        'description': f"{bc.FEATURE_TM_RNA} overlap with ({tmRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: CDS (%s/%s) [%i, %i] overlapping tmRNA (%s) [%i, %i], %s, contig=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], tmRNA['gene'], tmRNA['start'], tmRNA['stop'], overlap, cds['contig']
                    )
            # tRNA overlaps
            for tRNA in contig_t_rnas[contig['id']]:
                if(cds['stop'] < tRNA['start'] or cds['start'] > tRNA['stop']):
                    continue
                else:  # overlap -> remove cds
                    overlap = f"[{max(cds['start'], tRNA['start'])},{min(cds['stop'], tRNA['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_T_RNA,
                        'description': f"{bc.FEATURE_T_RNA} overlap with ({tRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: CDS (%s/%s) [%i, %i] overlapping tRNA (%s) [%i, %i], %s, contig=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], tRNA['gene'], tRNA['start'], tRNA['stop'], overlap, cds['contig']
                    )
            # rRNA overlaps
            for rRNA in contig_r_rnas[contig['id']]:
                if(cds['stop'] < rRNA['start'] or cds['start'] > rRNA['stop']):
                    continue
                else:  # overlap -> remove cds
                    overlap = f"[{max(cds['start'], rRNA['start'])},{min(cds['stop'], rRNA['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_R_RNA,
                        'description': f"{bc.FEATURE_R_RNA} overlap with ({rRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: CDS (%s/%s) [%i, %i] overlapping rRNA (%s) [%i, %i], %s, contig=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], rRNA['gene'], rRNA['start'], rRNA['stop'], overlap, cds['contig']
                    )
            # CRISPR overlaps
            for crispr in contig_crispr_arrays[contig['id']]:
                if(cds['stop'] < crispr['start'] or cds['start'] > crispr['stop']):
                    continue
                else:  # overlap -> remove cds
                    overlap = f"[{max(cds['start'], crispr['start'])},{min(cds['stop'], crispr['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_CRISPR,
                        'description': f'overlaps {bc.FEATURE_CRISPR} at {overlap}'
                    }
                    log.info(
                        "overlap: CDS (%s/%s) [%i, %i] overlapping CRISPR [%i, %i], %s, contig=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], crispr['start'], crispr['stop'], overlap, cds['contig']
                    )
        
        # remove sORF overlapping with tRNAs, tmRNAs, rRNAs, CRISPRs, inframe CDSs, shorter inframe sORFs
        for sorf in contig_sorfs[contig['id']]:
            # tmRNA overlaps
            for tmRNA in contig_tm_rnas[contig['id']]:
                if(sorf['stop'] < tmRNA['start'] or sorf['start'] > tmRNA['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], tmRNA['start'])},{min(sorf['stop'], tmRNA['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_TM_RNA,
                        'description': f"{bc.FEATURE_TM_RNA} overlap with ({tmRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping tRNA (%s) [%i, %i], %s, contig=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], tmRNA['gene'], tmRNA['start'], tmRNA['stop'], overlap, sorf['contig']
                    )
            # tRNA overlaps
            for tRNA in contig_t_rnas[contig['id']]:
                if(sorf['stop'] < tRNA['start'] or sorf['start'] > tRNA['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], tRNA['start'])},{min(sorf['stop'], tRNA['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_T_RNA,
                        'description': f"{bc.FEATURE_T_RNA} overlap with ({tRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping tRNA (%s) [%i, %i], %s, contig=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], tRNA['gene'], tRNA['start'], tRNA['stop'], overlap, sorf['contig']
                    )
            # rRNA overlaps
            for rRNA in contig_r_rnas[contig['id']]:
                if(sorf['stop'] < rRNA['start'] or sorf['start'] > rRNA['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], rRNA['start'])},{min(sorf['stop'], rRNA['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_R_RNA,
                        'description': f"{bc.FEATURE_R_RNA} overlap with ({rRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping rRNA (%s) [%i, %i], %s, contig=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], rRNA['gene'], rRNA['start'], rRNA['stop'], overlap, sorf['contig']
                    )
            # CRISPR overlaps
            for crispr in contig_crispr_arrays[contig['id']]:
                if(sorf['stop'] < crispr['start'] or sorf['start'] > crispr['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], crispr['start'])},{min(sorf['stop'], crispr['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_CRISPR,
                        'description': f'overlaps {bc.FEATURE_CRISPR} at {overlap}'
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping CRISPR [%i, %i], %s, contig=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], crispr['start'], crispr['stop'], overlap, sorf['contig']
                    )
            # CDS overlaps (skipped as most overlaps are filtered in sORF detection)
            
            # sORF overlaps
            for overlap_sorf in contig_sorfs[contig['id']]:
                if(sorf['stop'] < overlap_sorf['start'] or sorf['start'] > overlap_sorf['stop']):
                    continue  # no overlap
                elif(sorf['start'] == overlap_sorf['start'] and sorf['stop'] == overlap_sorf['stop']):
                    continue  # same
                else:  # overlap -> remove sorf
                    score_sorf = calc_sorf_annotation_score(sorf)
                    score_overlap_sorf = calc_sorf_annotation_score(overlap_sorf)

                    if(score_sorf < score_overlap_sorf):  # lower annotation score
                        overlap = f"[{max(sorf['start'], overlap_sorf['start'])},{min(sorf['stop'], overlap_sorf['stop'])}]"
                        sorf['discarded'] = {
                            'type': bc.DISCARD_TYPE_OVERLAP,
                            'feature_type': bc.FEATURE_SORF,
                            'description': f"overlaps {bc.FEATURE_SORF} ({overlap_sorf.get('gene', '-')}/{overlap_sorf.get('product', '-')}) at {overlap} with lower score ({score_sorf}/{score_overlap_sorf})"
                        }
                        log.info(
                            "discard overlapping sORF: (%s/%s) [%i, %i] overlapping sORF (%s/%s) [%i, %i], %s, contig=%s, lower annotation score (%i/%i)",
                            sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], overlap_sorf.get('gene', '-'), overlap_sorf.get('product', '-'), overlap_sorf['start'], overlap_sorf['stop'], overlap, sorf['contig'], score_sorf, score_overlap_sorf
                        )
                    elif(score_sorf == score_overlap_sorf and len(sorf['sequence']) < len(overlap_sorf['sequence'])):  # equal annotation score but shorter sequence -> potential fragment or too short ORF prediction
                        overlap = f"[{max(sorf['start'], overlap_sorf['start'])},{min(sorf['stop'], overlap_sorf['stop'])}]"
                        sorf['discarded'] = {
                            'type': bc.DISCARD_TYPE_OVERLAP,
                            'feature_type': bc.FEATURE_SORF,
                            'description': f"overlaps {bc.FEATURE_SORF} ({overlap_sorf.get('gene', '-')}/{overlap_sorf.get('product', '-')}) at {overlap} with equal score ({score_sorf}) but lower length ({len(sorf['sequence'])}/{len(overlap_sorf['sequence'])})"
                        }
                        log.info(
                            "discard overlapping sORF: (%s/%s) [%i, %i] overlapping sORF (%s/%s) [%i, %i], %s, contig=%s, equal annotation score (%i), lower length (%i/%i)",
                            sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], overlap_sorf.get('gene', '-'), overlap_sorf.get('product', '-'), overlap_sorf['start'], overlap_sorf['stop'], overlap, sorf['contig'], score_sorf, len(sorf['sequence']), len(overlap_sorf['sequence'])
                        )


def calc_sorf_annotation_score(sorf):
    """Calc an annotation score rewarding each identification & annotation"""
    score = 0

    if('ups' in sorf):
        score += 1
    
    ips = sorf.get('ips', None)
    if(ips):
        score += 1
        ips_gene = ips.get('gene', None)
        if(ips_gene):
            score += 1
        ips_product = ips.get('product', None)
        if(ips_product):
            score += 1
    
    psc = sorf.get('psc', None)
    if(psc):
        score += 1
        psc_gene = psc.get('gene', None)
        if(psc_gene):
            score += 1
        psc_product = psc.get('product', None)
        if(psc_product):
            score += 1
    return score




import argparse
import hashlib
import logging
import multiprocessing as mp
import os
import sys
import subprocess as sp

import bakta
import bakta.constants as bc
import bakta.config as cfg

log = logging.getLogger('UTILS')


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog='bakta',
        description='Rapid & standardized annotation of bacterial genomes & plasmids.',
        add_help=False
    )
    parser.add_argument('genome', metavar='<genome>', help='(Draft) genome in fasta format')

    arg_group_io = parser.add_argument_group('Input / Output')
    arg_group_io.add_argument('--db', '-d', action='store', default=None, help='Database path (default = <bakta_path>/db)')
    arg_group_io.add_argument('--min-contig-length', '-m', action='store', type=int, default=1, dest='min_contig_length', help='Minimum contig size (default = 1)')
    arg_group_io.add_argument('--prefix', '-p', action='store', default=None, help='Prefix for output files')
    arg_group_io.add_argument('--output', '-o', action='store', default=os.getcwd(), help='Output directory (default = current working directory)')

    arg_group_organism = parser.add_argument_group('Organism')
    arg_group_organism.add_argument('--genus', action='store', default=None, help='Genus name')
    arg_group_organism.add_argument('--species', action='store', default=None, help='Species name')
    arg_group_organism.add_argument('--strain', action='store', default=None, help='Strain name')
    arg_group_organism.add_argument('--plasmid', action='store', default=None, help='Plasmid name')
    
    arg_group_annotation = parser.add_argument_group('Annotation')
    arg_group_annotation.add_argument('--complete', action='store_true', help="All sequences are complete replicons (chromosome/plasmid[s])")
    arg_group_annotation.add_argument('--prodigal-tf', action='store', default=None, dest='prodigal_tf', help='Path to existing Prodigal training file to use for CDS prediction')
    arg_group_annotation.add_argument('--translation-table', action='store', type=int, default=11, choices=[11, 4], dest='translation_table', help='Translation table to use: 11/4 (default = 11)')
    arg_group_annotation.add_argument('--gram', action='store', default='?', choices=['+', '-', '?'], help="Gram type: +/-/? (default = '?')")
    arg_group_annotation.add_argument('--locus', action='store', default=None, help="Locus prefix (instead of 'contig')")
    arg_group_annotation.add_argument('--locus-tag', action='store', default=None, dest='locus_tag', help='Locus tag prefix')
    arg_group_annotation.add_argument('--keep-contig-headers', action='store_true', dest='keep_contig_headers', help='Keep original contig headers')
    arg_group_annotation.add_argument('--replicons', '-r', action='store', default=None, dest='replicons', help="Replicon information table (TSV)")

    arg_group_workflow = parser.add_argument_group('Workflow')
    arg_group_workflow.add_argument('--skip-trna', action='store_true', dest='skip_trna', help="Skip tRNA detection & annotation")
    arg_group_workflow.add_argument('--skip-tmrna', action='store_true', dest='skip_tmrna', help="Skip tmRNA detection & annotation")
    arg_group_workflow.add_argument('--skip-rrna', action='store_true', dest='skip_rrna', help="Skip rRNA detection & annotation")
    arg_group_workflow.add_argument('--skip-ncrna', action='store_true', dest='skip_ncrna', help="Skip ncRNA detection & annotation")
    arg_group_workflow.add_argument('--skip-ncrna-region', action='store_true', dest='skip_ncrna_region', help="Skip ncRNA region detection & annotation")
    arg_group_workflow.add_argument('--skip-crispr', action='store_true', dest='skip_crispr', help="Skip CRISPR array detection & annotation")
    arg_group_workflow.add_argument('--skip-cds', action='store_true', dest='skip_cds', help="Skip CDS detection & annotation")
    arg_group_workflow.add_argument('--skip-sorf', action='store_true', dest='skip_sorf', help="Skip sORF detection & annotation")
    arg_group_workflow.add_argument('--skip-gap', action='store_true', dest='skip_gap', help="Skip gap detection & annotation")
    arg_group_workflow.add_argument('--skip-ori', action='store_true', dest='skip_ori', help="Skip oriC/oriT detection & annotation")

    arg_group_general = parser.add_argument_group('General')
    arg_group_general.add_argument('--help', '-h', action='help', help='Show this help message and exit')
    arg_group_general.add_argument('--verbose', '-v', action='store_true', help='Print verbose information')
    arg_group_general.add_argument('--threads', '-t', action='store', type=int, default=mp.cpu_count(), help='Number of threads to use (default = number of available CPUs)')
    arg_group_general.add_argument('--tmp-dir', action='store', default=None, dest='tmp_dir', help='Location for temporary files (default = system dependent auto detection)')
    arg_group_general.add_argument('--version', action='version', version='%(prog)s ' + bakta.__version__)
    arg_group_general.add_argument('--citation', action='store_true', help='Print citation')
    return parser.parse_args()


def test_dependencies():
    """Test the proper installation of necessary 3rd party executables."""

    # test tRNAscan-SE
    if(cfg.skip_trna is False):
        try:
            sp.check_call(
                ['tRNAscan-SE', '-h'],
                stdout=sp.DEVNULL,
                stderr=sp.DEVNULL
            )
        except FileNotFoundError:
            log.exception('tRNAscan-SE not found!')
            sys.exit('ERROR: tRNAscan-SE not executable! Please either install tRNAscan-SE (>=v2.0.6) or skip the annotation of tRNAs via \'--skip-trna\'.')
        except:
            pass

    # test Aragorn
    if(cfg.skip_tmrna is False):
        try:
            sp.check_call(
                ['aragorn', '-h'],
                stdout=sp.DEVNULL,
                stderr=sp.DEVNULL
            )
        except FileNotFoundError:
            log.exception('aragorn not found!')
            sys.exit('ERROR: aragorn not executable! Please either install Aragorn (>=1.2.38) or skip the annotation of tmRNAs via \'--skip-tmrna\'.')
        except:
            pass

    # test cmscan
    if(cfg.skip_rrna is False or cfg.skip_ncrna is False or cfg.skip_ncrna_region is False):
        try:
            sp.check_call(
                ['cmscan', '-h'],
                stdout=sp.DEVNULL,
                stderr=sp.DEVNULL
            )
        except FileNotFoundError:
            log.exception('cmscan not found!')
            sys.exit('ERROR: \'cmscan\' not executable! Please either install Infernal (>=v1.1.2) or skip the annotation of rRNAs, ncRNAs & ncRNA regions via \'--skip-rrna --skip-ncrna --skip-ncrna-region\'.')
        except:
            pass

    # test prodigal
    if(cfg.skip_cds is False):
        try:
            sp.check_call(
                ['prodigal', '-v'],
                stdout=sp.DEVNULL,
                stderr=sp.DEVNULL
            )
        except FileNotFoundError:
            log.exception('prodigal not found!')
            sys.exit('ERROR: \'prodigal\' not executable! Please either install Prodigal (>=v2.6.3) or skip the annotation of CDSs via \'--skip-cds\'.')
        except:
            pass

    # test hmmsearch
    if(cfg.skip_cds is False or cfg.skip_sorf is False):
        try:
            sp.check_call(
                ['hmmsearch', '-h'],
                stdout=sp.DEVNULL,
                stderr=sp.DEVNULL
            )
        except FileNotFoundError:
            log.exception('hmmsearch not found!')
            sys.exit('ERROR: \'hmmsearch\' not executable! Please either install HMMER (>=3.3.1) or skip the annotation of CDSs & sORFs via \'--skip-cds --skip-sorf\'.')
        except:
            pass

    # test diamond
    if(cfg.skip_cds is False or cfg.skip_sorf is False):
        try:
            sp.check_call(
                ['diamond', '--version'],
                stdout=sp.DEVNULL,
                stderr=sp.DEVNULL
            )
        except FileNotFoundError:
            log.exception('diamond not found!')
            sys.exit('ERROR: \'diamond\' not executable! Please either install Diamond (>=v2.0.4) or skip the annotation of CDSs & sORFs via \'--skip-cds --skip-sorf\'.')
        except:
            pass

    # test pilercr
    if(cfg.skip_crispr is False):
        try:
            sp.check_call(
                ['pilercr', '-options'],
                stdout=sp.DEVNULL,
                stderr=sp.DEVNULL
            )
        except FileNotFoundError:
            log.exception('pilercr not found!')
            sys.exit('ERROR: pilercr not executable! Please either install PILER-CR or skip the annotation of CRISPRs via \'--skip-crispr\'.')
        except:
            pass


def create_locus_prefix(contigs):
    """Create either genus/species or sequence MD5 hex based locus prefix.
    Max locus name length is 37 for GenBank -> 32 + _ + 4 digits"""
    if(cfg.genus != '' and cfg.species != ''):
        locus_prefix = cfg.genus[:1] + cfg.species[:2]
        return locus_prefix.upper()
    else:
        hash = hashlib.md5()
        for contig in contigs:
            hash.update(str.encode(contig['sequence']))
        return hash.hexdigest()[0:5]


def create_locus_tag_prefix(contigs):
    """Create either genus/species or sequence MD5 hex based locus tag prefix."""
    hash = hashlib.md5()
    for contig in contigs:
        hash.update(str.encode(contig['sequence']))
    hexdigest = hash.hexdigest().upper()
    locus_prefix = []
    i = 0
    while i < 6:
        c = hexdigest[i]
        if(c >= '0' and c <= '9'):
            c = chr(ord('F') + int(c) + 1) 
        locus_prefix.append(c)
        i += 1
    return ''.join(locus_prefix)


def calc_aa_hash(seq):
    aa_hash = hashlib.md5(seq.encode('utf-8'))
    return (aa_hash.digest(), aa_hash.hexdigest())


def has_annotation(feature, attribute):
    value = feature.get(attribute, None)
    if(value is not None and value != ''):
        return True
    else:
        return False


def calc_genome_stats(genome, features):
    
    genome_size = genome['size']
    log.info('genome-size=%i', genome_size)

    # N50
    gc_sum = 0
    n_sum = 0
    n50 = 0
    contig_length_sum = 0
    for contig in genome['contigs']:
        seq = contig['sequence']
        gc_sum += seq.count('G') + seq.count('C')
        n_sum += seq.count('N')
        contig_length = len(seq)
        contig_length_sum += contig_length
        if(contig_length_sum >= genome_size / 2):
            n50 = contig_length
            break
    genome['n50'] = n50
    log.info('N50=%i', n50)

    gc_ratio = gc_sum / (genome_size - n_sum)
    genome['gc'] = gc_ratio
    log.info('GC=%0.3f', gc_ratio)
    
    n_ratio = n_sum / genome_size
    genome['n_ratio'] = n_ratio
    log.info('N=%0.3f', n_ratio)

    contigs_by_id = {c['id']: c for c in genome['contigs']}
    coding_nts = 0
    for feat in features:
        if(feat.get('edge', False)):
            sequence_length = contigs_by_id[feat['contig']]['length']
            coding_nts += feat['stop'] + (sequence_length - feat['start'] + 1)  # feature coding nucleotides
        else:
            coding_nts += feat['stop'] - feat['start'] + 1  # feature coding nucleotides
    coding_ratio = coding_nts / (genome_size - n_sum)
    genome['coding_ratio'] = coding_ratio
    log.info('coding-ratio=%0.3f', coding_ratio)
    
    return {
        'gc': gc_ratio,
        'n_ratio': n_ratio,
        'n50': n50,
        'coding_ratio': coding_ratio
    }


def parse_replicon_table(replicon_table_path):
    replicons = {}
    try:
        with replicon_table_path.open() as fh:
            for line in fh:
                (original_locus_id, new_locus_id, replicon_type, topology, name) = line.strip().split('\t')
                # TODO: add locus id checks
                if(new_locus_id == '' or new_locus_id == ''):
                    new_locus_id = None
                replicon_type = replicon_type.lower()
                if(replicon_type == 'c' or 'chrom' in replicon_type):
                    replicon_type = bc.REPLICON_CHROMOSOME
                elif(replicon_type == 'p' or 'plasmid' in replicon_type):
                    replicon_type = bc.REPLICON_PLASMID
                else:
                    replicon_type = bc.REPLICON_CONTIG
                topology = topology.lower()
                if(topology == 'c' or 'circ' in topology):
                    topology = bc.TOPOLOGY_CIRCULAR
                else:
                    topology = bc.TOPOLOGY_LINEAR
                if(replicon_type == bc.REPLICON_CONTIG):
                    topology == bc.TOPOLOGY_LINEAR
                if(name == '' or name == '-'):
                    name = None
                replicon = {
                    'original_locus_id': original_locus_id,
                    'new_locus_id': new_locus_id,
                    'replicon_type': replicon_type,
                    'topology': topology,
                    'name': name
                }
                log.info(
                    'replicon info: orig-id=%s, new-id=%s, type=%s, topology=%s, name=%s',
                    replicon['original_locus_id'], replicon['new_locus_id'], replicon['replicon_type'], replicon['topology'], replicon['name']
                )
                replicons[original_locus_id] = replicon
    except:
        log.error('wrong replicon table format!')
        sys.exit('ERROR: wrong replicon table file format!')
    return replicons


def qc_contigs(contigs, replicons):
    valid_contigs = []
    contig_counter = 1
    contig_prefix = cfg.locus if cfg.locus else 'contig'
    organism_definition = f"[organism={cfg.taxon}]" if cfg.taxon else None
    
    complete_genome = True
    plasmid_number = 1
    for contig in contigs:
        if(contig['length'] >= cfg.min_contig_length):
            contig_id_generated = f'{contig_prefix}_{contig_counter}'
            contig['simple_id'] = contig_id_generated
            contig_counter += 1

            if('circular=true' in contig['description'].lower()):  # detection of Unicycler circularized sequences
                contig['complete'] = True
                contig['topology'] = bc.TOPOLOGY_CIRCULAR
                log.debug('qc: detected Unicycler circular topology via description: id=%s, description=%s', contig['id'], contig['description'])
            if('complete' in contig['description'].lower()):
                contig['complete'] = True
                contig['topology'] = bc.TOPOLOGY_CIRCULAR
                log.debug('qc: detected complete replicon via description: id=%s, description=%s', contig['id'], contig['description'])
            if('chromosome' in contig['description'].lower()):
                contig['type'] = bc.REPLICON_CHROMOSOME
                log.debug('qc: detected chromosome replicon type via description: id=%s, description=%s', contig['id'], contig['description'])
            elif('plasmid' in contig['description'].lower()):
                contig['type'] = bc.REPLICON_PLASMID
                log.debug('qc: detected plasmid replicon type via description: id=%s, description=%s', contig['id'], contig['description'])
            
            if(not cfg.keep_contig_headers):
                contig['orig_id'] = contig['id']
                contig['id'] = contig_id_generated
                contig['orig_description'] = contig['description']
                contig_desc = []
                if(organism_definition):
                    contig_desc.append(organism_definition)
                if(cfg.strain):
                    contig_desc.append(f'[strain={cfg.strain}]')
                if(cfg.complete or contig['complete']):
                    contig_desc.append('[completeness=complete]')
                    if(contig['topology'] != bc.REPLICON_CONTIG):
                        contig_desc.append(f"[topology={contig['topology']}]")
                contig['description'] = ' '.join(contig_desc)
            if(cfg.complete):
                contig['complete'] = True
                contig['topology'] = bc.TOPOLOGY_CIRCULAR
                if(contig['length'] >= bc.REPLICON_LENGTH_THRESHOLD_CHROMOSOME):
                    contig['type'] = bc.REPLICON_CHROMOSOME
                elif(contig['length'] < bc.REPLICON_LENGTH_THRESHOLD_PLASMID):
                    contig['type'] = bc.REPLICON_PLASMID
            
            valid_contigs.append(contig)
            
            if(replicons):  # use user provided replicon table
                contig_id = contig['orig_id'] if 'orig_id' in contig else contig['id']
                replicon = replicons.get(contig_id, None)
                if(replicon):
                    contig['type'] = replicon['replicon_type']
                    contig['topology'] = replicon['topology']
                    if(replicon['name']):
                        contig['name'] = replicon['name']
                    if(replicon['replicon_type'] != bc.REPLICON_CONTIG):
                        contig['complete'] = True
                    if(not cfg.keep_contig_headers):
                        contig['id'] = replicon['new_locus_id'] if replicon['new_locus_id'] else contig['simple_id']
                        if(replicon['replicon_type'] != bc.REPLICON_CONTIG and 'completeness' not in contig['description']):
                            contig['description'] += ' [completeness=complete]'
                        if('topology' not in contig['description']):
                            contig['description'] += f" [topology={replicon['topology']}]"
                        if(replicon['replicon_type'] == bc.REPLICON_PLASMID and replicon['name']):
                            contig['description'] += f" [plasmid-name={replicon['name']}]"
                    contig.pop('simple_id')

            if(contig['complete'] and contig['type'] == bc.REPLICON_PLASMID and not contig.get('name', None)):
                contig['name'] = f'unnamed{plasmid_number}'
                plasmid_number += 1
            
            if(contig['type'] == bc.REPLICON_CONTIG):
                complete_genome = False
            
            log.info(
                "qc: revised sequence: id=%s, orig-id=%s, type=%s, complete=%s, topology=%s, name=%s, description='%s', orig-description='%s'",
                contig['id'], contig.get('orig_id', ''), contig['type'], contig['complete'], contig['topology'], contig.get('name', ''), contig['description'], contig.get('orig_description', '')
            )
    return valid_contigs, complete_genome



import os
from pathlib import Path
from subprocess import run


def test_genome():
    # test genome arguments

    # no provided
    proc = run(["bin/bakta", '--db', 'test/db'])
    assert proc.returncode != 0

    # non-existing
    proc = run(["bin/bakta", '--db', 'test/db', 'foo.fasta'])
    assert proc.returncode != 0


def test_database(tmpdir):
    # test database arguments

    # no provided
    proc = run(["bin/bakta", 'test/data/NC_002127.1.fna'])
    assert proc.returncode != 0

    # parameter missing path
    proc = run(["bin/bakta", '--db', 'test/data/NC_002127.1.fna'])
    assert proc.returncode != 0

    # parameter wrong path
    proc = run(["bin/bakta", '--db', 'test/', 'test/data/NC_002127.1.fna'])
    assert proc.returncode != 0

    # parameter OK
    proc = run(["bin/bakta", '--db', 'test/db', '--output', tmpdir, '--skip-tmrna', '--skip-trna', '--skip-rrna', '--skip-ncrna', '--skip-ncrna-region', '--skip-crispr', '--skip-cds', '--skip-sorf', '--skip-ori', '--skip-gap', 'test/data/NC_002127.1.fna'])
    assert proc.returncode == 0

    # environment missing path
    env = os.environ
    env['BAKTA_DB'] = ''
    proc = run(["bin/bakta", 'test/data/NC_002127.1.fna'], env=env)
    assert proc.returncode != 0

    # parameter wrong path
    env['BAKTA_DB'] = 'test/'
    proc = run(["bin/bakta", 'test/data/NC_002127.1.fna'], env=env)
    assert proc.returncode != 0

    # parameter OK
    env['BAKTA_DB'] = 'test/db'
    proc = run(["bin/bakta", '--output', tmpdir, 'test/data/NC_002127.1.fna'], env=env)
    assert proc.returncode == 0


def test_prodigal_tf(tmpdir):
    # test prodigal training file arguments

    # missing
    proc = run(["bin/bakta", '--db', 'test/db', '--output', tmpdir, '--prefix', 'test', '--prodigal-tf', 'test/data/NC_002127.1.fna'])
    assert proc.returncode != 0

    # non-existing
    proc = run(["bin/bakta", '--db', 'test/db', '--output', tmpdir, '--prefix', 'test', '--prodigal-tf', 'test/data/foo', 'test/data/NC_002127.1.fna'])
    assert proc.returncode != 0

    # OK
    proc = run(["bin/bakta", '--db', 'test/db', '--output', tmpdir, '--prefix', 'test', '--prodigal-tf', 'test/data/prodigal.tf', 'test/data/NC_002127.1.fna'])
    assert proc.returncode == 0
    assert Path.exists(Path(tmpdir).joinpath('test.log'))
    assert Path.exists(Path(tmpdir).joinpath('test.json'))
    assert Path.exists(Path(tmpdir).joinpath('test.tsv'))
    assert Path.exists(Path(tmpdir).joinpath('test.gff3'))
    assert Path.exists(Path(tmpdir).joinpath('test.gbff'))
    assert Path.exists(Path(tmpdir).joinpath('test.fna'))
    assert Path.exists(Path(tmpdir).joinpath('test.faa'))



import json
import logging
import os
import sys

import bakta
import bakta.config as cfg

log = logging.getLogger('DB')


def check():
    """Check if database directory exists, is accessible and contains necessary files."""

    if(cfg.db_path is None):
        log.error('directory not provided nor detected!')
        sys.exit('ERROR: database directory not provided nor detected! Please provide a valid path to the database directory.')

    if(not os.access(str(cfg.db_path), os.R_OK & os.X_OK)):
        log.error('directory (%s) not readable/accessible!', cfg.db_path)
        sys.exit(f'ERROR: database directory ({cfg.db_path}) not readable/accessible!')

    version_path = cfg.db_path.joinpath('version.json')
    if(not os.access(str(version_path), os.R_OK) or not version_path.is_file()):
        log.error('version file not readable!')
        sys.exit('ERROR: database version file (db.yaml) not readable!')
    
    try:
        with version_path.open() as fh:
            db_info = json.load(fh)
    except:
        log.exception('could not parse database version file!')
        sys.exit('ERROR: could not parse database version file!')
    
    for key in ['date', 'major', 'minor']:
        if(key not in db_info):
            log.error('wrong db version info file content! missed key=%s', key)
            sys.exit(f"ERROR: wrong db version info file format! Missed key '{key}' in JSON structure.")
    
    log.info('detected: major=%i', db_info['major'])
    log.info('detected: minor=%i', db_info['minor'])
    log.info('detected: date=%s', db_info['date'])
    if(db_info['major'] < bakta.__db_schema_version__):
        log.error('wrong database version detected! required=%i, detected=%i', bakta.__db_schema_version__, db_info['major'])
        sys.exit(f"ERROR: wrong database version detected!\nBakta version {bakta.__version__} requires database version {bakta.__db_schema_version__}.x, but {db_info['major']}.{db_info['minor']} was detected. Please, update the database from https://doi.org/10.5281/zenodo.4247253")
    elif(db_info['major'] > bakta.__db_schema_version__):
        log.error('wrong database version detected! required=%i, detected=%i', bakta.__db_schema_version__, db_info['major'])
        sys.exit(f"ERROR: wrong database version detected!\nBakta version {bakta.__version__} requires database version {bakta.__db_schema_version__}.x, but {db_info['major']}.{db_info['minor']} was detected. Please, update Bakta or download a compatible database version from https://doi.org/10.5281/zenodo.4247253")

    file_names = [
        'antifam.h3f',
        'antifam.h3i',
        'antifam.h3m',
        'antifam.h3p',
        'bakta.db',
        'ncRNA-genes.i1f',
        'ncRNA-genes.i1i',
        'ncRNA-genes.i1m',
        'ncRNA-genes.i1p',
        'ncRNA-regions.i1f',
        'ncRNA-regions.i1i',
        'ncRNA-regions.i1m',
        'ncRNA-regions.i1p',
        'oric.fna',
        'orit.fna',
        'psc.dmnd',
        'rfam-go.tsv',
        'rRNA.i1f',
        'rRNA.i1i',
        'rRNA.i1m',
        'rRNA.i1p',
        'sorf.dmnd'
    ]

    for file_name in file_names:
        path = cfg.db_path.joinpath(file_name)
        if(not os.access(str(path), os.R_OK) or not path.is_file()):
            log.error('file not readable! file=%s', file_name)
            sys.exit(f'ERROR: database file ({file_name}) not readable!')

    return db_info
